/***************************************************************************************************
Module: Handle Viewer.c
***************************************************************************************************/


#include <ntddk.h>
#include "HandleViewer.h"


////////////////////////////////////////////////////////////////////////////////////////////////////


#define DEVICE_NAME					L"\\Device\\HandleViewer"
#define SYMBOLIC_LINK				L"\\DosDevices\\HandleViewer"
#define GLOBAL_SYMBOLIC_LINK		L"\\DosDevices\\Global\\HandleViewer"


////////////////////////////////////////////////////////////////////////////////////////////////////


#define IOCTL_VIEW_HANDLE \
	CTL_CODE(FILE_DEVICE_UNKNOWN, 0X800, METHOD_BUFFERED, FILE_ANY_ACCESS)


////////////////////////////////////////////////////////////////////////////////////////////////////


UNICODE_STRING g_SymbolicLink;
ULONG_PTR g_HandleTableOffset;


////////////////////////////////////////////////////////////////////////////////////////////////////


BOOLEAN CheckSystemVersion()
{
	BOOLEAN Ret = FALSE;
	ULONG ulBuildNumber = 0;
	RTL_OSVERSIONINFOW	osi;

	osi.dwOSVersionInfoSize = sizeof(RTL_OSVERSIONINFOW);
	RtlFillMemory(&osi, sizeof(RTL_OSVERSIONINFOW), 0);
	RtlGetVersion(&osi);
	ulBuildNumber = osi.dwBuildNumber;

	switch (ulBuildNumber)
	{
		// Win7
		case 7600:
		{
			g_HandleTableOffset = 0x200;
			Ret = TRUE;
			break;
		}
		// Win7 SP1
		case 7601:
		{
			g_HandleTableOffset = 0x200;
			Ret = TRUE;
			break;
		}
		// Win8 / WinSer2012
		case 9200:
		{
			g_HandleTableOffset = 0x0;
			Ret = TRUE;
			break;
		}
		// Win8.1 / WinSer2012 R2
		case 9600:
		{
			g_HandleTableOffset = 0x0;
			Ret = TRUE;
			break;
		}
		// Win10 10240
		case 10240:
		{
			g_HandleTableOffset = 0x418;
			Ret = TRUE;
			break;
		}
		// Win10 10586
		case 10586:
		{
			g_HandleTableOffset = 0x418;
			Ret = TRUE;
			break;
		}
		default:
			break;
	}

	return Ret;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


PHANDLE_TABLE_ENTRY ExpLookupHandleTableEntry(
	IN PHANDLE_TABLE HandleTable,
	IN EXHANDLE tHandle
)
{
	ULONG_PTR i, j, k;
	ULONG_PTR CapturedTable;
	ULONG TableLevel;
	PHANDLE_TABLE_ENTRY Entry = NULL;
	EXHANDLE Handle;

	PUCHAR TableLevel1;
	PUCHAR TableLevel2;
	PUCHAR TableLevel3;

	ULONG_PTR MaxHandle;

	PAGED_CODE();

	Handle = tHandle;
	Handle.TagBits = 0;

	MaxHandle = *(volatile ULONG *)&HandleTable->NextHandleNeedingPool;
	if (Handle.Value >= MaxHandle)
	{
		return NULL;
	}

	CapturedTable = *(volatile ULONG_PTR *)&HandleTable->TableCode;
	TableLevel = (ULONG)(CapturedTable & LEVEL_CODE_MASK);
	CapturedTable = CapturedTable - TableLevel;

	switch (TableLevel)
	{
		case 0:
		{
			TableLevel1 = (PUCHAR)CapturedTable;

			Entry = (PHANDLE_TABLE_ENTRY)&TableLevel1[Handle.Value *
				(sizeof(HANDLE_TABLE_ENTRY) / HANDLE_VALUE_INC)];

			break;
		}

		case 1:
		{
			TableLevel2 = (PUCHAR)CapturedTable;

			i = Handle.Value % (LOWLEVEL_COUNT * HANDLE_VALUE_INC);
			Handle.Value -= i;
			j = Handle.Value / ((LOWLEVEL_COUNT * HANDLE_VALUE_INC) / sizeof(PHANDLE_TABLE_ENTRY));

			TableLevel1 = (PUCHAR) *(PHANDLE_TABLE_ENTRY *)&TableLevel2[j];
			Entry = (PHANDLE_TABLE_ENTRY)&TableLevel1[i * (sizeof(HANDLE_TABLE_ENTRY) / HANDLE_VALUE_INC)];

			break;
		}

		case 2:
		{
			TableLevel3 = (PUCHAR)CapturedTable;

			i = Handle.Value % (LOWLEVEL_COUNT * HANDLE_VALUE_INC);
			Handle.Value -= i;
			k = Handle.Value / ((LOWLEVEL_COUNT * HANDLE_VALUE_INC) / sizeof(PHANDLE_TABLE_ENTRY));
			j = k % (MIDLEVEL_COUNT * sizeof(PHANDLE_TABLE_ENTRY));
			k -= j;
			k /= MIDLEVEL_COUNT;

			TableLevel2 = (PUCHAR) *(PHANDLE_TABLE_ENTRY *)&TableLevel3[k];
			TableLevel1 = (PUCHAR) *(PHANDLE_TABLE_ENTRY *)&TableLevel2[j];
			Entry = (PHANDLE_TABLE_ENTRY)&TableLevel1[i * (sizeof(HANDLE_TABLE_ENTRY) / HANDLE_VALUE_INC)];

			break;
		}

		default: _assume(0);
	}

	return Entry;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


NTSTATUS ViewHandle(ULONG32 ProcessId, POBJECT_INFO Buffer)
{
	PEPROCESS EProcess = NULL;
	ULONG_PTR Handle = 0;
	PHANDLE_TABLE_ENTRY Entry = NULL;
	PVOID Object = NULL;
	POBJECT_TYPE ObjectType = NULL;

	if (!NT_SUCCESS(PsLookupProcessByProcessId((HANDLE)ProcessId, &EProcess)))
	{
		return STATUS_UNSUCCESSFUL;
	}

	for (Handle = 0;; Handle += HANDLE_VALUE_INC)
	{
		Entry = ExpLookupHandleTableEntry(*(PHANDLE_TABLE*)((PUCHAR)EProcess + g_HandleTableOffset), *(PEXHANDLE)&Handle);
		if (Entry == NULL)
		{
			break;
		}

		*(ULONG_PTR*)&Object = Entry->ObjectPointerBits;
		*(ULONG_PTR*)&Object <<= 4;
		if (Object == NULL)
		{
			continue;
		}

		*(ULONG_PTR*)&Object |= 0xFFFF000000000000;
		*(ULONG_PTR*)&Object += 0x30;
		ObjectType = ObGetObjectType(Object);
		if (ObjectType == NULL)
		{
			continue;
		}

		wcscpy(Buffer->szTypeName, *(PCWSTR*)((PUCHAR)ObjectType + 0x18));
		Buffer->Handle = (HANDLE)Handle;
		Buffer->AccessMask = Entry->GrantedAccessBits;
		Buffer->Address = Object;

		Buffer++;
	}

	ObDereferenceObject(EProcess);

	return STATUS_SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


NTSTATUS DispatchCreate(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	UNREFERENCED_PARAMETER(DeviceObject);

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


NTSTATUS DispatchClose(PDEVICE_OBJECT DeviceObject, PIRP pIrp)
{
	UNREFERENCED_PARAMETER(DeviceObject);

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


NTSTATUS DispatchDeviceControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
	PIO_STACK_LOCATION IoStackLocation = NULL;
	PVOID SystemBuffer = NULL;
	ULONG IoControlCode = 0;
	ULONG InputBufferLength = 0;
	ULONG OutputBufferLength = 0;
	ULONG_PTR RetLength = 0;
	NTSTATUS Status = STATUS_SUCCESS;
	ULONG32 ProcessId = 0;

	UNREFERENCED_PARAMETER(DeviceObject);

	IoStackLocation = IoGetCurrentIrpStackLocation(Irp);
	IoControlCode = IoStackLocation->Parameters.DeviceIoControl.IoControlCode;
	InputBufferLength = IoStackLocation->Parameters.DeviceIoControl.InputBufferLength;
	OutputBufferLength = IoStackLocation->Parameters.DeviceIoControl.OutputBufferLength;
	SystemBuffer = Irp->AssociatedIrp.SystemBuffer;

	switch (IoControlCode)
	{
		case IOCTL_VIEW_HANDLE:
		{
			if (InputBufferLength != sizeof(ULONG32))
			{
				RetLength = 0;
				Status = STATUS_INVALID_PARAMETER;
				break;
			}

			ProcessId = *(PULONG32)SystemBuffer;
			RtlZeroMemory(SystemBuffer, OutputBufferLength);
			Status = ViewHandle(ProcessId, (POBJECT_INFO)SystemBuffer);
			RetLength = OutputBufferLength;

			break;
		}

		default:
			break;
	}

	Irp->IoStatus.Status = Status;
	Irp->IoStatus.Information = RetLength;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return Status;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


void DriverUnload(PDRIVER_OBJECT DriverObject)
{
	IoDeleteSymbolicLink(&g_SymbolicLink);
	IoDeleteDevice(DriverObject->DeviceObject);
}


////////////////////////////////////////////////////////////////////////////////////////////////////


NTSTATUS CreateDevice(PDRIVER_OBJECT DriverObject)
{
	NTSTATUS Status = STATUS_SUCCESS;
	UNICODE_STRING DeviceName;
	PDEVICE_OBJECT DeviceObject;

	RtlInitUnicodeString(&DeviceName, DEVICE_NAME);
	Status = IoCreateDevice(DriverObject, 0, &DeviceName, FILE_DEVICE_UNKNOWN, 0, FALSE, &DeviceObject);
	if (!NT_SUCCESS(Status))
	{
		return Status;
	}

	// Using "Global DosDevice Directory" or not
	if (IoIsWdmVersionAvailable(1, 0x10))
	{
		// Windows 2000 or later
		RtlInitUnicodeString(&g_SymbolicLink, GLOBAL_SYMBOLIC_LINK);
	}
	else
	{
		// Windows Me/98
		RtlInitUnicodeString(&g_SymbolicLink, SYMBOLIC_LINK);
	}

	Status = IoCreateSymbolicLink(&g_SymbolicLink, &DeviceName);
	if (!NT_SUCCESS(Status))
	{
		IoDeleteDevice(DeviceObject);
		return Status;
	}

	return Status;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


NTSTATUS DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegisterPath)
{
	NTSTATUS Status = STATUS_SUCCESS;

	UNREFERENCED_PARAMETER(RegisterPath);

	CheckSystemVersion();

	DriverObject->MajorFunction[IRP_MJ_CREATE] = DispatchCreate;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = DispatchClose;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DispatchDeviceControl;
	DriverObject->DriverUnload = DriverUnload;

	Status = CreateDevice(DriverObject);
	if (!NT_SUCCESS(Status))
	{
		return Status;
	}

	return Status;
}


//////////////////////////////////////////// End Of File ///////////////////////////////////////////