/***************************************************************************************************
Module: Handle Viewer.h
***************************************************************************************************/


#pragma once


////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef HANDLE_VIEWER_H
#define HANDLE_VIEWER_H


////////////////////////////////////////////////////////////////////////////////////////////////////


#pragma warning(disable:4200)
#pragma warning(disable:4201)
#pragma warning(disable:4214)


////////////////////////////////////////////////////////////////////////////////////////////////////


#define HANDLE_VALUE_INC 4

#define TABLE_PAGE_SIZE PAGE_SIZE
#define LOWLEVEL_COUNT (TABLE_PAGE_SIZE / sizeof(HANDLE_TABLE_ENTRY))
#define MIDLEVEL_COUNT (PAGE_SIZE / sizeof(PHANDLE_TABLE_ENTRY))

#define LEVEL_CODE_MASK 3


////////////////////////////////////////////////////////////////////////////////////////////////////


typedef struct _OBJECT_INFO
{
	WCHAR szTypeName[20];
	HANDLE Handle;
	ULONG32 AccessMask;
	PVOID Address;
} OBJECT_INFO, *POBJECT_INFO;


////////////////////////////////////////////////////////////////////////////////////////////////////


typedef struct _EXHANDLE
{
	union
	{
		struct
		{
			ULONG32 TagBits : 2;
			ULONG32 Index : 30;
		};
		HANDLE GenericHandleOverlay;
		ULONG_PTR Value;
	};
} EXHANDLE, *PEXHANDLE;


////////////////////////////////////////////////////////////////////////////////////////////////////


//1: kd> dt nt!_HANDLE_TABLE_ENTRY
//+ 0x000 VolatileLowValue : Int8B
//+ 0x000 LowValue : Int8B
//+ 0x000 InfoTable : Ptr64 _HANDLE_TABLE_ENTRY_INFO
//+ 0x008 HighValue : Int8B
//+ 0x008 NextFreeHandleEntry : Ptr64 _HANDLE_TABLE_ENTRY
//+ 0x008 LeafHandleValue : _EXHANDLE
//+ 0x000 RefCountField : Int8B
//+ 0x000 Unlocked : Pos 0, 1 Bit
//+ 0x000 RefCnt : Pos 1, 16 Bits
//+ 0x000 Attributes : Pos 17, 3 Bits
//+ 0x000 ObjectPointerBits : Pos 20, 44 Bits
//+ 0x008 GrantedAccessBits : Pos 0, 25 Bits
//+ 0x008 NoRightsUpgrade : Pos 25, 1 Bit
//+ 0x008 Spare1 : Pos 26, 6 Bits
//+ 0x00c Spare2 : Uint4B

typedef struct _HANDLE_TABLE_ENTRY
{
	union
	{
		LONG_PTR VolatileLowValue;
		LONG_PTR LowValue;
		PVOID InfoTable;
		LONG_PTR RefCountField;
		struct
		{
			ULONG_PTR Unlocked : 1;
			ULONG_PTR RefCnt : 16;
			ULONG_PTR Attributes : 3;
			ULONG_PTR ObjectPointerBits : 44;
		};
	};
	union
	{
		LONG_PTR HighValue;
		struct _HANDLE_TABLE_ENTRY *NextFreeHandleEntry;
		EXHANDLE LeafHandleValue;
		struct
		{
			ULONG32 GrantedAccessBits : 25;
			ULONG32 NoRightsUpgrade : 1;
			ULONG32 Spare1 : 6;
		};
		ULONG32 Spare2;
	};
} HANDLE_TABLE_ENTRY, *PHANDLE_TABLE_ENTRY;

//typedef struct _HANDLE_TABLE_ENTRY
//{
//	union
//	{
//		PVOID Object;
//		ULONG32 ObAttributes;
//		PVOID InfoTable;
//		ULONG_PTR Value;
//	};
//	union
//	{
//		ULONG32 GranteAccess;
//		struct
//		{
//			UINT16 GranteAccessIndex;
//			UINT16 CreatorBackTraceIndex;
//		};
//		ULONG32 NextFreeTableEntry;
//	};
//} HANDLE_TABLE_ENTRY, *PHANDLE_TABLE_ENTRY;


////////////////////////////////////////////////////////////////////////////////////////////////////


//1: kd> dt nt!_HANDLE_TABLE_FREE_LIST
//+ 0x000 FreeListLock     : _EX_PUSH_LOCK
//+ 0x008 FirstFreeHandleEntry : Ptr64 _HANDLE_TABLE_ENTRY
//+ 0x010 LastFreeHandleEntry : Ptr64 _HANDLE_TABLE_ENTRY
//+ 0x018 HandleCount : Int4B
//+ 0x01c HighWaterMark : Uint4B
//+ 0x020 Reserved : [8] Uint4B

typedef struct _HANDLE_TABLE_FREE_LIST
{
	ULONG_PTR FreeListLock;
	PHANDLE_TABLE_ENTRY FirstFreeHandleEntry;
	PHANDLE_TABLE_ENTRY lastFreeHandleEntry;
	LONG32 HandleCount;
	ULONG32 HighWaterMark;
	ULONG32 Reserved[8];
} HANDLE_TABLE_FREE_LIST, *PHANDLE_TABLE_FREE_LIST;


////////////////////////////////////////////////////////////////////////////////////////////////////


//1: kd> dt nt!_HANDLE_TABLE
//+ 0x000 NextHandleNeedingPool : Uint4B
//+ 0x004 ExtraInfoPages : Int4B
//+ 0x008 TableCode : Uint8B
//+ 0x010 QuotaProcess : Ptr64 _EPROCESS
//+ 0x018 HandleTableList : _LIST_ENTRY
//+ 0x028 UniqueProcessId : Uint4B
//+ 0x02c Flags : Uint4B
//+ 0x02c StrictFIFO : Pos 0, 1 Bit
//+ 0x02c EnableHandleExceptions : Pos 1, 1 Bit
//+ 0x02c Rundown : Pos 2, 1 Bit
//+ 0x02c Duplicated : Pos 3, 1 Bit
//+ 0x02c RaiseUMExceptionOnInvalidHandleClose : Pos 4, 1 Bit
//+ 0x030 HandleContentionEvent : _EX_PUSH_LOCK
//+ 0x038 HandleTableLock : _EX_PUSH_LOCK
//+ 0x040 FreeLists : [1] _HANDLE_TABLE_FREE_LIST
//+ 0x040 ActualEntry : [32] UChar
//+ 0x060 DebugInfo : Ptr64 _HANDLE_TRACE_DEBUG_INFO

typedef struct _HANDLE_TABLE
{
	ULONG32 NextHandleNeedingPool;
	LONG32 ExtraInfoPages;
	ULONG_PTR TableCode;
	PEPROCESS QuotaProcess;
	LIST_ENTRY HandleTableList;
	ULONG32 UniqueProcessId;
	union
	{
		ULONG32 Flags;
		struct
		{
			BOOLEAN StrictFIFO : 1;
			BOOLEAN EnableHandleExceptions : 1;
			BOOLEAN Rundown : 1;
			BOOLEAN Duplicated : 1;
			BOOLEAN RaiseUMExceptionOnInvalidHandleClose : 1;
		};
	};
	ULONG_PTR HandleContentionEvent;
	ULONG_PTR HandleTableLock;
	union
	{
		HANDLE_TABLE_FREE_LIST FreeLists[1];
		BOOLEAN ActualEntry[32];
	};
	PVOID DebugInfo;
} HANDLE_TABLE, *PHANDLE_TABLE;

//typedef struct _HANDLE_TABLE
//{
//	ULONG_PTR TableCode;
//	PEPROCESS QuotaProcess;
//	HANDLE UniqueProcessId;
//	ULONG_PTR HandleLock;  // EX_PUSH_LOCK
//	LIST_ENTRY HandleTableList;
//	ULONG_PTR HandleContentionEvent;  // EX_PUSH_LOCK
//	PVOID DebugInfo;  // PHANDLE_TRACE_DEBUG_INFO
//	LONG32 ExtraInfoPages;
//	union
//	{
//		ULONG32 Flags;
//		BOOLEAN StrictFIFO : 1;
//	};
//	ULONG32 FirstFreeHandle;
//	PHANDLE_TABLE_ENTRY LastFreeHandleEntry;
//	ULONG32 HandleCount;
//	ULONG32 NextHandleNeedingPool;
//	ULONG32 HandleCountHighWatermark;
//} HANDLE_TABLE, *PHANDLE_TABLE;


////////////////////////////////////////////////////////////////////////////////////////////////////


NTKERNELAPI NTSTATUS PsLookupProcessByProcessId(
	__in HANDLE ProcessId,
	__deref_out PEPROCESS *Process
);


NTKERNELAPI POBJECT_TYPE ObGetObjectType(PVOID Object);


////////////////////////////////////////////////////////////////////////////////////////////////////


#endif


//////////////////////////////////////////// End Of File ///////////////////////////////////////////