/***************************************************************************************************
Module: DriverControl.c
Author: ZChameleon @ 2015
***************************************************************************************************/


#include <Windows.h>
#include "DriverControl.h"


////////////////////////////////////////////////////////////////////////////////////////////////////


BOOL DrvCtl_Install(PCTSTR szServiceName, PCTSTR szDriverPath)
{
	HANDLE hDriverFile;
	SC_HANDLE schSCManager;
	SC_HANDLE schService;
	
	hDriverFile = CreateFile(szDriverPath, GENERIC_READ,
		FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hDriverFile == INVALID_HANDLE_VALUE)
	{
		DbgOut("CreateFile:%d", GetLastError());
		return FALSE;
	}
	CloseHandle(hDriverFile);

	schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (schSCManager == NULL)
	{
		DbgOut("OpenSCManager:%d", GetLastError());
		return FALSE;
	}

	schService = CreateService(schSCManager, szServiceName, szServiceName,
		SERVICE_ALL_ACCESS, SERVICE_KERNEL_DRIVER, SERVICE_DEMAND_START,
		SERVICE_ERROR_NORMAL, szDriverPath, NULL, NULL, NULL, NULL, NULL);
	if (schService == NULL)
	{
		DbgOut("CreateService:%d", GetLastError());
		CloseServiceHandle(schSCManager);
		return FALSE;
	}

	CloseServiceHandle(schService);
	CloseServiceHandle(schSCManager);

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


BOOL DrvCtl_Uninstall(PCTSTR szServiceName)
{
	SC_HANDLE schSCManager;
	SC_HANDLE schService;

	schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (schSCManager == NULL)
	{
		DbgOut("OpenSCManager:%d", GetLastError());
		return FALSE;
	}

	schService = OpenService(schSCManager, szServiceName, DELETE);
	if (schService == NULL)
	{
		DbgOut("OpenService:%d", GetLastError());
		CloseServiceHandle(schSCManager);
		return FALSE;
	}

	if (!DeleteService(schService))
	{
		DbgOut("DeleteService:%d", GetLastError());
		return FALSE;
	}

	CloseServiceHandle(schService);
	CloseServiceHandle(schSCManager);

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


BOOL DrvCtl_Start(PCTSTR szServiceName)
{
	SC_HANDLE schSCManager;
	SC_HANDLE schService;

	schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (schSCManager == NULL)
	{
		DbgOut("OpenSCManager:%d", GetLastError());
		return FALSE;
	}

	schService = OpenService(schSCManager, szServiceName, SERVICE_ALL_ACCESS);
	if (schService == NULL)
	{
		DbgOut("OpenService:%d", GetLastError());
		CloseServiceHandle(schSCManager);
		return FALSE;
	}

	if (!StartService(schService, 0, NULL))
	{
		DbgOut("StartService:%d", GetLastError());
		CloseServiceHandle(schService);
		CloseServiceHandle(schSCManager);
		return FALSE;
	}

	CloseServiceHandle(schService);
	CloseServiceHandle(schSCManager);

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


BOOL DrvCtl_Stop(PCTSTR szServiceName)
{
	SC_HANDLE schSCManager;
	SC_HANDLE schService;
	SERVICE_STATUS ss;

	schSCManager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
	if (schSCManager == NULL)
	{
		DbgOut("OpenSCManager:%d", GetLastError());
		return FALSE;
	}

	schService = OpenService(schSCManager, szServiceName,
		SERVICE_STOP | SERVICE_QUERY_STATUS | SERVICE_ENUMERATE_DEPENDENTS);
	if (schService == NULL)
	{
		DbgOut("OpenService:%d", GetLastError());
		CloseServiceHandle(schSCManager);
		return FALSE;
	}

	if (!ControlService(schService, SERVICE_CONTROL_STOP, &ss))
	{
		DbgOut("ControlService:%d", GetLastError());
		CloseServiceHandle(schService);
		CloseServiceHandle(schSCManager);
		return FALSE;
	}

	CloseServiceHandle(schService);
	CloseServiceHandle(schSCManager);

	return TRUE;
}


////////////////////////////////////////////////////////////////////////////////////////////////////


BOOL DrvCtl_Control(PCTSTR szSymbolicLink, DWORD IoControlCode,
	PVOID InBuffer, DWORD InBufferSize, PVOID OutBuffer, DWORD OutBufferSize)
{
	HANDLE hDevice = INVALID_HANDLE_VALUE;
	DWORD Junk = 0;

	hDevice = CreateFile(szSymbolicLink, 0, FILE_SHARE_READ |
		FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
	if (hDevice == INVALID_HANDLE_VALUE)
	{
		DbgOut("CreateFile:%d", GetLastError());
		return FALSE;
	}

	if (!DeviceIoControl(hDevice, IoControlCode, InBuffer,
		InBufferSize, OutBuffer, OutBufferSize, &Junk, NULL))
	{
		DbgOut("DeviceIoControl:%d", GetLastError());
		CloseHandle(hDevice);
		return FALSE;
	}

	CloseHandle(hDevice);

	return TRUE;
}


//////////////////////////////////////////// End Of File ///////////////////////////////////////////